from collections import defaultdict

import requests
import yaml
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.db.models import Max
from teklia_toolbox.config import ConfigParser, ConfigurationError

from arkindex.process.models import ArkindexFeature, FeatureUsage, Worker, WorkerType, WorkerVersion, WorkerVersionState

REQUEST_TIMEOUT = (30, 60)
WORKER_YAML_VERSION = 2


def parse_config():
    parser = ConfigParser()
    parser.add_option("version", type=str)


    features_parser = parser.add_subparser("features")
    for feature in ArkindexFeature:
        feature_parser = features_parser.add_subparser(feature.value, allow_extra_keys=False, default={})
        feature_parser.add_option("image", type=str, default=None)
        feature_parser.add_option("command", type=str, default=None)
        teklia_worker_parser = feature_parser.add_subparser("teklia_worker", allow_extra_keys=False, default=None)
        teklia_worker_parser.add_option("name", type=str, default=None)
        teklia_worker_parser.add_option("version", type=str, default=None)
        teklia_worker_parser.add_option("slug", type=str, default=None)

    parsed = parser.parse(settings.BASE_DIR / "system_workers.yml")

    errors = defaultdict(list)
    for feature, config in parsed["features"].items():
        if config["image"] and config["teklia_worker"]:
            errors[feature].append("Exactly one of image/command or teklia_parser must be set")
            continue
        if (config["command"] and config["image"] is None):
            errors[feature].append("command argument must be set with the image argument")
        if (subparser := config["teklia_worker"]) and (None in subparser.values()):
            errors[feature].append("teklia_parser configuration must define a name, a version and a slug")
    if errors:
        raise ConfigurationError(errors)
    return parsed


class Command(BaseCommand):
    help = "Update the workers used to provide Arkindex features to the versions compatible with this release."

    def get_system_worker(self, feature: ArkindexFeature, repo = None) -> Worker:
        """
        Update or create a `system` worker for this feature.
        Creates the `system` worker type if it does not exist.
        """
        try:
            conflicting_worker = Worker.objects.exclude(type__slug="system").filter(slug=feature.value).get()
            self.stdout.write(f"Worker {conflicting_worker.name} ({conflicting_worker.id}) uses the same slug with a different worker type.")
            self.stdout.write("Please update this worker's type to `system` or change its slug.")
            raise CommandError(f"Cannot create system worker with slug {feature.value} due to a conflict.")
        except Worker.DoesNotExist:
            pass

        worker_type, created = WorkerType.objects.get_or_create(
            slug="system",
            defaults={"display_name": "System"},
        )

        if created:
            self.stdout.write(f"Created new System worker type ({worker_type.id})")

        worker, created = Worker.objects.get_or_create(
            type=worker_type,
            slug=feature.value,
            repository_url=repo,
            defaults={
                "name": feature.name,
                "public": True,
            },
        )

        if created:
            self.stdout.write(f"Created new {worker.name} system worker")
        else:
            self.stdout.write(f"Using existing system worker {worker.name}")
            self. update_existing_worker(worker)

        return worker

    def update_existing_worker(self, worker: Worker) -> None:
        """
        Check that an existing worker has the correct attributes to have one of its versions provide a feature
        """
        updated = False

        if worker.archived is not None:
            self.stdout.write("Unarchiving worker")
            worker.archived = None
            updated = True

        if not worker.public:
            self.stdout.write("Marking worker as public")
            worker.public = True
            updated = True

        if updated:
            worker.save()
        else:
            self.stdout.write("Worker is up to date")

    def update_or_create_version(self, worker: Worker, feature: ArkindexFeature, docker_image: str, docker_command: str = None, configuration: dict = {}) -> None:
        """
        On a specified worker, assigns an existing version to a feature or creates a new one.
        Expects that no version is already assigned to this feature on any worker.
        """
        assert worker.archived is None, "Cannot assign a version on an archived worker to a feature"

        # Since there can be worker versions using the same Docker image but with different commands,
        # we need to filter by command as well as by Docker image.
        docker_command_filter = {"configuration__docker__command": docker_command} if docker_command else {"configuration__docker__command__isnull": True}
        # docker_image_iid is not unique, so we sort by state and take the most recent one.
        # Sorting by state means we prefer `available` versions first, then `created`, then `error`.
        worker_version = worker.versions.filter(
            docker_image_iid=docker_image,
            gpu_usage=FeatureUsage.Disabled,
            model_usage=FeatureUsage.Disabled,
            configuration=configuration,
            **docker_command_filter
        ).order_by("state", "-updated").first()

        if worker_version:
            self.stdout.write(f"Assigning existing worker version {worker_version.id} to the feature")
            worker_version.feature = feature

            if docker_command is not None:
                if "docker" not in worker_version.configuration:
                    worker_version.configuration["docker"] = {}
                worker_version.configuration["docker"]["command"] = docker_command
                self.stdout.write(f"Setup docker command: {docker_command}")

            if worker_version.state != WorkerVersionState.Available:
                self.stdout.write("Marking the worker version as available")
                worker_version.state = WorkerVersionState.Available

            worker_version.save()
            self.stdout.write(self.style.SUCCESS(f"Using existing worker version {worker_version.id}"))
            return

        self.stdout.write("Creating new worker version")
        max_version = worker.versions.aggregate(max_version=Max("version"))["max_version"] or 0
        worker_version = worker.versions.create(
            docker_image_iid=docker_image,
            feature=feature,
            state=WorkerVersionState.Available,
            version=max_version + 1,
            configuration=configuration,
        )
        self.stdout.write(self.style.SUCCESS(f"Using new worker version {worker_version.id}"))

    def check_existing_version(self, worker_version: WorkerVersion, feature: ArkindexFeature, *, docker_image: str, docker_command: str = None, configuration: dict = {}, repo: str = None) -> None:
        self.stdout.write(f"Current worker version: {worker_version.id} ({worker_version.docker_image_iid})")
        valid = True

        if worker_version.docker_image_iid != docker_image:
            self.stdout.write(self.style.WARNING("This version has an invalid Docker image."))
            valid = False

        if worker_version.worker.archived is not None:
            self.stdout.write(self.style.WARNING("This version is part of an archived worker."))
            valid = False

        if worker_version.docker_command is None:
            if docker_command is not None:
                self.stdout.write(self.style.WARNING(f"This version requires the following Docker command: {docker_command}."))
                valid = False

        else:
            if docker_command is not None and docker_command == worker_version.docker_command:
                self.stdout.write("This version uses the required Docker command.")
            else:
                self.stdout.write(self.style.WARNING("This version uses a custom Docker command which could interfere with the feature."))
                valid = False

        if worker_version.configuration != configuration:
            self.stdout.write(self.style.WARNING(
                "This version uses a custom configuration which could interfere with the feature."
            ))
            valid = False

        if valid:
            # Ensure the worker is public, and keep using this worker version
            self.update_existing_worker(worker_version.worker)
            self.stdout.write(self.style.SUCCESS(f"Worker version for {feature.name} is up to date"))
            return

        self.stdout.write("Unassigning feature from the current version")
        worker_version.feature = None
        worker_version.save(update_fields=["feature"])

        # Create a new version on the existing worker, even if it isn't a system worker
        worker = worker_version.worker
        if worker.archived is not None:
            # except if it is archived, since the new version would be invalid
            worker = self.get_system_worker(feature, repo)
        else:
            self.update_existing_worker(worker_version.worker)

        self.update_or_create_version(worker, feature, docker_image, docker_command, configuration)

    @transaction.atomic
    def update_feature(self, feature: ArkindexFeature, *, image, command, configuration = {}, repo = None):
        self.stdout.write(f"Using {image} to provide {feature.name}")

        try:
            worker_version = WorkerVersion.objects.get_by_feature(feature)
            self.check_existing_version(
                worker_version,
                feature,
                docker_image=image,
                docker_command=command,
                configuration=configuration
            )
        except WorkerVersion.DoesNotExist:
            worker = self.get_system_worker(feature, repo)
            self.update_or_create_version(worker, feature, image, command, configuration)

    def update_feature_from_worker(self, feature, *, name, version, slug):
        """
        Update a feature from a worker repository hosted on https://gitlab.teklia.com/
        """
        repo = f"https://gitlab.teklia.com/{name}"
        self.stdout.write(f"Configuring feature {feature} from {repo}")

        # Retrieve the .arkindex.yml file with no auth
        url = f"{repo}/-/raw/{version}/.arkindex.yml"
        with requests.get(url, timeout=REQUEST_TIMEOUT) as resp:
            try:
                resp.raise_for_status()
            except requests.exceptions.HTTPError as e:
                raise CommandError(f"Error retrieving configuration at {url}: {e.response.status_code}.")
            data = yaml.safe_load(resp.content)
        if not isinstance(data, dict) or data.get("version", 0) < WORKER_YAML_VERSION or not data.get("workers"):
            raise CommandError(f"Error retrieving configuration at {url}: invalid YAML configuration.")
        # Look for the worker matching feature's slug
        worker_conf = next((worker for worker in data["workers"] if worker["slug"] == slug), None)
        if worker_conf is None:
            raise CommandError(f"No worker with slug {slug} in .arkindex.yml at {url}.")
        image = f"registry.gitlab.teklia.com/{name}:{version}"
        command = worker_conf.get("docker", {}).get("command", None)
        self.update_feature(feature, image=image, command=command, configuration=worker_conf, repo=repo)

    def handle(self, *args, **options):
        config = parse_config()
        for feature_value, feature_config in config["features"].items():
            feature = ArkindexFeature(feature_value)
            self.stdout.write(f" {feature.name} ".center(80, "─"))
            worker = feature_config.pop("teklia_worker", None)
            if worker is not None:
                self.update_feature_from_worker(feature, **worker)
            else:
                configuration = {}
                if (command := feature_config["command"]):
                    configuration["docker"] = {"command": command}
                self.update_feature(feature, **feature_config, configuration=configuration)
