from unittest.mock import call, patch
from uuid import uuid4

from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import State
from arkindex.process.models import ProcessMode, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestProcessClear(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            element_type=cls.corpus.types.get(slug="page")
        )
        cls.process.worker_runs.create(
            version=WorkerVersion.objects.get(worker__slug="reco"),
        )
        cls.process.worker_runs.create(
            version=WorkerVersion.objects.get(worker__slug="dla"),
        )

    def test_clear(self):
        self.assertEqual(self.process.worker_runs.count(), 2)
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.delete(reverse("api:clear-process", kwargs={"pk": str(self.process.id)}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.process.refresh_from_db()
        self.assertEqual(self.process.worker_runs.count(), 0)

    def test_requires_login(self):
        self.assertEqual(self.process.worker_runs.count(), 2)

        with self.assertNumQueries(0):
            response = self.client.delete(reverse("api:clear-process", kwargs={"pk": str(self.process.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.assertEqual(self.process.worker_runs.count(), 2)
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(reverse("api:clear-process", kwargs={"pk": str(self.process.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_does_not_exist(self):
        self.assertEqual(self.process.worker_runs.count(), 2)

        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:clear-process", kwargs={"pk": str(uuid4())}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_not_unscheduled(self):
        """
        Cannot clear a process that has already started
        """
        self.assertEqual(self.process.worker_runs.count(), 2)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.process.tasks.update(state=State.Running)
        self.assertEqual(self.process.state, State.Running)
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:clear-process", kwargs={"pk": str(self.process.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"__all__": ["A process can only be cleared before getting started."]})

    def test_unscheduled(self):
        """
        Cannot clear a process that has tasks, even unscheduled
        """
        self.assertEqual(self.process.worker_runs.count(), 2)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process.run()
        self.assertEqual(self.process.state, State.Unscheduled)
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:clear-process", kwargs={"pk": str(self.process.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"__all__": ["A process can only be cleared before getting started."]})

    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Contributor.value)
    def test_requires_permissions(self, get_max_level_mock):
        self.assertEqual(self.process.worker_runs.count(), 2)
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(reverse("api:clear-process", kwargs={"pk": str(self.process.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have a sufficient access level to this process."})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    def test_wrong_process_mode(self):
        """
        This endpoint is only supported for Workers and Dataset processes
        """
        self.client.force_login(self.user)

        cases = set(ProcessMode) - {ProcessMode.Dataset, ProcessMode.Workers, ProcessMode.Local}
        for mode in cases:
            with self.subTest(mode=mode):
                self.process.mode = mode
                self.process.save()

                with self.assertNumQueries(4):
                    response = self.client.delete(reverse("api:clear-process", kwargs={"pk": str(self.process.id)}))
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), {"__all__": ["Only Workers and Dataset processes can be cleared."]})

    def test_local_process_mode(self):
        """
        This endpoint is only supported for Workers and Dataset processes
        """
        self.client.force_login(self.user)
        local_process = self.user.processes.filter(mode=ProcessMode.Local).first()
        with self.assertNumQueries(3):
            response = self.client.delete(reverse("api:clear-process", kwargs={"pk": str(local_process.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})
