from unittest.mock import call, patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.images.models import ImageServer
from arkindex.ponos.models import Farm
from arkindex.process.models import (
    ArkindexFeature,
    Process,
    ProcessMode,
    Worker,
    WorkerType,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import Role


class TestCreateS3Import(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.import_worker_version = WorkerVersion.objects.create(
            worker=Worker.objects.create(
                name="S3 Import",
                slug="s3_import",
                type=WorkerType.objects.get(slug="import"),
            ),
            version=1,
            state=WorkerVersionState.Available,
            feature=ArkindexFeature.S3Ingest,
            docker_image_iid="registry.gitlab.teklia.com/arkindex/s4:latest",
            configuration={
                "docker": {
                    "command": "rm -rf /",
                },
            },
        )
        ImageServer.objects.create(id=999, display_name="Ingest image server", url="https://dev.null.teklia.com")
        cls.corpus.types.create(slug="folder", display_name="Folder", folder=True)

    @override_settings(ARKINDEX_FEATURES={"ingest": False})
    def test_requires_feature_flag(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:s3-import-create"), {
                "corpus_id": str(self.corpus.id),
                "folder_type": "volume",
                "element_type": "page",
                "bucket_name": "blah",
                "prefix": "a/b/c",
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), ["S3 ingest is not available on this instance."])

    @override_settings(ARKINDEX_FEATURES={"ingest": True})
    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:s3-import-create"), {
                "corpus_id": str(self.corpus.id),
                "folder_type": "volume",
                "element_type": "page",
                "bucket_name": "blah",
                "prefix": "a/b/c",
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"ingest": True})
    def test_requires_verified(self):
        self.user.can_ingest = True
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:s3-import-create"), {
                "corpus_id": str(self.corpus.id),
                "folder_type": "volume",
                "element_type": "page",
                "bucket_name": "blah",
                "prefix": "a/b/c",
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"ingest": True})
    def test_requires_scope(self):
        self.assertFalse(self.user.can_ingest)
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:s3-import-create"), {
                "corpus_id": str(self.corpus.id),
                "folder_type": "volume",
                "element_type": "page",
                "bucket_name": "blah",
                "prefix": "a/b/c",
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"ingest": True})
    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_invalid(self, filter_rights_mock):
        filter_rights_mock.return_value = Corpus.objects.filter(id=self.corpus.id)
        self.user.can_ingest = True
        self.user.save()
        self.client.force_login(self.user)
        private_corpus = Corpus.objects.create(name="nope")
        private_element = private_corpus.elements.create(type=private_corpus.types.create(slug="something"), name="can't touch this")
        cases = [
            (
                {},
                {
                    "bucket_name": ["This field is required."],
                    "corpus_id": ["This field is required."],
                }
            ),
            (
                {"corpus_id": "", "element_id": "", "bucket_name": "", "folder_type": "", "element_type": ""},
                {
                    "bucket_name": ["This field may not be blank."],
                    "corpus_id": ["This field may not be null."],
                },
            ),
            (
                {"corpus_id": "lol", "element_id": "lol", "folder_type": "a", "element_type": "a", "bucket_name": "a"},
                {
                    "bucket_name": ["Ensure this field has at least 3 characters."],
                    "corpus_id": ["“lol” is not a valid UUID."],
                    "element_id": ["“lol” is not a valid UUID."],
                },
            ),
            (
                {"corpus_id": str(self.corpus.id), "bucket_name": "a" * 100, "prefix": "/" * 2000},
                {
                    "bucket_name": ["Ensure this field has no more than 63 characters."],
                    "prefix": ["Ensure this field has no more than 1024 characters."],
                },
            ),
            (
                {"corpus_id": str(self.corpus.id), "bucket_name": "aaa", "folder_type": "not_found", "element_type": "not_found"},
                {
                    "element_type": [f"Corpus {self.corpus.id} does not have a 'not_found' type, or it is a folder type"],
                    "folder_type": [f"Corpus {self.corpus.id} does not have a 'not_found' type, or it is not a folder type"],
                },
            ),
            (
                {"corpus_id": str(private_corpus.id), "bucket_name": "bucket_nem"},
                {"corpus_id": [f'Invalid pk "{private_corpus.id}" - object does not exist.']},
            ),
            (
                {"corpus_id": str(self.corpus.id), "element_id": str(private_element.id), "bucket_name": "lol"},
                {"element_id": [f'Invalid pk "{private_element.id}" - object does not exist.']},
            ),
        ]
        for request, expected in cases:
            filter_rights_mock.reset_mock()
            with self.subTest(request=request):
                response = self.client.post(reverse("api:s3-import-create"), request)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected)
                self.assertEqual(filter_rights_mock.call_count, 1)
                self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Admin.value))

    @override_settings(
        ARKINDEX_FEATURES={"ingest": True},
        PONOS_DEFAULT_ENV={},
        INGEST_IMAGESERVER_ID=999,
        INGEST_S3_ENDPOINT="http://s3.null.teklia.com",
        INGEST_S3_REGION=None,
        INGEST_S3_ACCESS_KEY="🔑",
        INGEST_S3_SECRET_KEY="its-secret-i-wont-tell-you",
        INGEST_PREFIX_BY_BUCKET_NAME=True,
    )
    def test_create(self):
        self.user.can_ingest = True
        self.user.save()
        self.client.force_login(self.user)
        element = self.corpus.elements.get(name="Volume 1")
        farm = Farm.objects.create(name="Crypto farm")

        with self.assertNumQueries(24):
            response = self.client.post(reverse("api:s3-import-create"), {
                "corpus_id": str(self.corpus.id),
                "element_id": str(element.id),
                "folder_type": "volume",
                "element_type": "page",
                "bucket_name": "blah",
                "prefix": "a/b/c",
                "farm_id": str(farm.id),
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        data = response.json()
        process = Process.objects.get(id=data["id"])

        self.assertDictEqual(data, {"id": str(process.id)})
        self.assertEqual(process.mode, ProcessMode.S3)
        self.assertEqual(process.creator, self.user)
        self.assertEqual(process.corpus_id, self.corpus.id)
        self.assertEqual(process.element_id, element.id)
        self.assertEqual(process.folder_type, self.corpus.types.get(slug="volume"))
        self.assertEqual(process.element_type, self.corpus.types.get(slug="page"))
        self.assertEqual(process.bucket_name, "blah")
        self.assertEqual(process.prefix, "a/b/c")
        self.assertEqual(process.farm, farm)

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.import_worker_version)
        self.assertListEqual(worker_run.parents, [])
        self.assertEqual(worker_run.configuration.name, f"Configuration for process {process.id}")
        self.assertDictEqual(worker_run.configuration.configuration, {
            "bucket": "blah",
            "bucket_prefix": True,
            "iiif_base_url": "https://dev.null.teklia.com",
        })
        self.assertIsNone(worker_run.model_version_id)

        task = process.tasks.get()
        self.assertEqual(task.slug, "import_s3")
        self.assertEqual(task.image, "registry.gitlab.teklia.com/arkindex/s4:latest")
        self.assertEqual(task.command, "rm -rf /")
        self.assertDictEqual(task.env, {
            "ARKINDEX_CORPUS_ID": str(self.corpus.id),
            "ARKINDEX_PROCESS_ID": str(process.id),
            "ARKINDEX_TASK_TOKEN": task.token,
            "ARKINDEX_WORKER_RUN_ID": str(worker_run.id),
            "INGEST_S3_ENDPOINT": "http://s3.null.teklia.com",
            "INGEST_S3_ACCESS_KEY": "🔑",
            "INGEST_S3_SECRET_KEY": "its-secret-i-wont-tell-you",
        })

    @override_settings(
        ARKINDEX_FEATURES={"ingest": True},
        PONOS_DEFAULT_ENV={},
        INGEST_IMAGESERVER_ID=999,
        INGEST_S3_ENDPOINT="http://s3.null.teklia.com",
        INGEST_S3_REGION=None,
        INGEST_S3_ACCESS_KEY="🔑",
        INGEST_S3_SECRET_KEY="its-secret-i-wont-tell-you",
        INGEST_PREFIX_BY_BUCKET_NAME=True,
    )
    def test_defaults(self):
        self.user.can_ingest = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(22):
            response = self.client.post(reverse("api:s3-import-create"), {
                "corpus_id": str(self.corpus.id),
                "bucket_name": "blah",
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data["id"])

        self.assertDictEqual(data, {"id": str(process.id)})
        self.assertEqual(process.mode, ProcessMode.S3)
        self.assertEqual(process.creator, self.user)
        self.assertEqual(process.farm, None)
        self.assertEqual(process.corpus_id, self.corpus.id)
        self.assertIsNone(process.element_id)
        self.assertEqual(process.folder_type, self.corpus.types.get(slug="folder"))
        self.assertEqual(process.element_type, self.corpus.types.get(slug="page"))
        self.assertEqual(process.bucket_name, "blah")
        self.assertIsNone(process.prefix)

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.import_worker_version)
        self.assertListEqual(worker_run.parents, [])
        self.assertEqual(worker_run.configuration.name, f"Configuration for process {process.id}")
        self.assertDictEqual(worker_run.configuration.configuration, {
            "bucket": "blah",
            "bucket_prefix": True,
            "iiif_base_url": "https://dev.null.teklia.com",
        })
        self.assertIsNone(worker_run.model_version_id)

        task = process.tasks.get()
        self.assertEqual(task.slug, "import_s3")
        self.assertEqual(task.image, "registry.gitlab.teklia.com/arkindex/s4:latest")
        self.assertEqual(task.command, "rm -rf /")
        self.assertDictEqual(task.env, {
            "ARKINDEX_CORPUS_ID": str(self.corpus.id),
            "ARKINDEX_PROCESS_ID": str(process.id),
            "ARKINDEX_TASK_TOKEN": task.token,
            "ARKINDEX_WORKER_RUN_ID": str(worker_run.id),
            "INGEST_S3_ENDPOINT": "http://s3.null.teklia.com",
            "INGEST_S3_ACCESS_KEY": "🔑",
            "INGEST_S3_SECRET_KEY": "its-secret-i-wont-tell-you",
        })

    @override_settings(
        ARKINDEX_FEATURES={"ingest": True},
        INGEST_IMAGESERVER_ID=999,
    )
    @patch("arkindex.ponos.models.Farm.is_available", return_value=False)
    def test_farm_guest(self, is_available_mock):
        self.user.can_ingest = True
        self.user.save()
        self.client.force_login(self.user)

        farm = Farm.objects.create(name="Crypto farm")

        with self.assertNumQueries(4):
            response = self.client.post(reverse("api:s3-import-create"), {
                "corpus_id": str(self.corpus.id),
                "bucket_name": "blah",
                "farm_id": str(farm.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "farm_id": ["You do not have access to this farm."],
        })
        self.assertFalse(Process.objects.filter(mode=ProcessMode.S3).exists())

        self.assertEqual(is_available_mock.call_count, 1)
        self.assertEqual(is_available_mock.call_args, call(self.user))

    @override_settings(
        ARKINDEX_FEATURES={"ingest": True},
        INGEST_IMAGESERVER_ID=999,
    )
    @patch("arkindex.process.serializers.ingest.get_default_farm")
    def test_default_farm(self, get_default_farm_mock):
        self.user.can_ingest = True
        self.user.save()
        self.client.force_login(self.user)

        default_farm = Farm.objects.create(name="Crypto farm")
        get_default_farm_mock.return_value = default_farm

        with self.assertNumQueries(22):
            response = self.client.post(reverse("api:s3-import-create"), {
                "corpus_id": str(self.corpus.id),
                "bucket_name": "blah",
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        process = Process.objects.get(id=response.json()["id"])
        self.assertEqual(process.farm, default_farm)

    @override_settings(
        ARKINDEX_FEATURES={"ingest": True},
        INGEST_IMAGESERVER_ID=999,
    )
    @patch("arkindex.ponos.models.Farm.is_available", return_value=False)
    @patch("arkindex.process.serializers.ingest.get_default_farm")
    def test_default_farm_guest(self, get_default_farm_mock, is_available_mock):
        self.user.can_ingest = True
        self.user.save()
        self.client.force_login(self.user)

        default_farm = Farm.objects.create(name="Crypto farm")
        get_default_farm_mock.return_value = default_farm

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:s3-import-create"), {
                "corpus_id": str(self.corpus.id),
                "bucket_name": "blah",
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "farm_id": ["You do not have access to this farm."],
        })
        self.assertFalse(Process.objects.filter(mode=ProcessMode.S3).exists())

        self.assertEqual(is_available_mock.call_count, 1)
        self.assertEqual(is_available_mock.call_args, call(self.user))
